@@ -2,6 +2,7 @@
  * OpenConnect (SSL + DTLS) VPN client
  *
  * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -25,7 +26,9 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
+#include <string.h>
 #include <ctype.h>
+#include <arpa/inet.h>
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
@@ -34,7 +37,7 @@
 
 /*
  * Data packets are encapsulated in the SSL stream as follows:
- * 
+ *
  * 0000: Magic "STF\x1"
  * 0004: Big-endian 16-bit length (not including 8-byte header)
  * 0006: Byte packet type (see openconnect.h)
@@ -77,14 +80,22 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	/* Clear old options which will be overwritten */
 	vpninfo->vpn_addr = vpninfo->vpn_netmask = NULL;
 	vpninfo->cstp_options = vpninfo->dtls_options = NULL;
+	vpninfo->vpn_domain = vpninfo->vpn_proxy_pac = NULL;
+
 	for (i=0; i<3; i++)
 		vpninfo->vpn_dns[i] = vpninfo->vpn_nbns[i] = NULL;
 
-	for (inc = vpninfo->split_includes; inc; inc = inc->next) {
+	for (inc = vpninfo->split_includes; inc; ) {
 		struct split_include *next = inc->next;
 		free(inc);
 		inc = next;
 	}
+	for (inc = vpninfo->split_excludes; inc; ) {
+		struct split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	vpninfo->split_includes = vpninfo->split_excludes = NULL;
  retry:
 	openconnect_SSL_printf(vpninfo->https_ssl, "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
 	openconnect_SSL_printf(vpninfo->https_ssl, "Host: %s\r\n", vpninfo->hostname);
@@ -101,14 +112,15 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	openconnect_SSL_printf(vpninfo->https_ssl, "X-DTLS-Master-Secret: ");
 	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++)
 		openconnect_SSL_printf(vpninfo->https_ssl, "%02X", vpninfo->dtls_secret[i]);
-	openconnect_SSL_printf(vpninfo->https_ssl, "\r\nX-DTLS-CipherSuite: AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA\r\n\r\n");
+	openconnect_SSL_printf(vpninfo->https_ssl, "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
+			       vpninfo->dtls_ciphers?:"AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
 
 	if (openconnect_SSL_gets(vpninfo->https_ssl, buf, 65536) < 0) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTPS response\n");
 		if (!retried) {
 			retried = 1;
 			openconnect_close_https(vpninfo);
-		
+
 			if (openconnect_open_https(vpninfo)) {
 				vpninfo->progress(vpninfo, PRG_ERR,
 						  "Failed to open HTTPS connection to %s\n",
@@ -124,7 +136,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		if (!strncmp(buf, "HTTP/1.1 503 ", 13)) {
 			/* "Service Unavailable. Why? */
 			char *reason = "<unknown>";
-			while ((i=openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+			while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
 				if (!strncmp(buf, "X-Reason: ", 10)) {
 					reason = buf + 10;
 					break;
@@ -134,7 +146,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 					  reason);
 			return -EINVAL;
 		}
-		vpninfo->progress(vpninfo, PRG_ERR, 
+		vpninfo->progress(vpninfo, PRG_ERR,
 				  "Got inappropriate HTTP CONNECT response: %s\n",
 				  buf);
 		if (!strncmp(buf, "HTTP/1.1 401 ", 13))
@@ -148,7 +160,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	/* We may have advertised it, but we only do it if the server agrees */
 	vpninfo->deflate = 0;
 
-	while ((i=openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
 		struct vpn_option *new_option;
 		char *colon = strchr(buf, ':');
 		if (!colon)
@@ -197,7 +209,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			if (!strcmp(colon, "deflate"))
 				vpninfo->deflate = 1;
 			else {
-				vpninfo->progress(vpninfo, PRG_ERR, 
+				vpninfo->progress(vpninfo, PRG_ERR,
 					"Unknown CSTP-Content-Encoding %s\n",
 					colon);
 				return -EINVAL;
@@ -226,6 +238,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			}
 		} else if (!strcmp(buf + 7, "Default-Domain")) {
 			vpninfo->vpn_domain = new_option->value;
+		} else if (!strcmp(buf + 7, "MSIE-Proxy-PAC-URL")) {
+			vpninfo->vpn_proxy_pac = new_option->value;
 		} else if (!strcmp(buf + 7, "Split-Include")) {
 			struct split_include *inc = malloc(sizeof(*inc));
 			if (!inc)
@@ -233,6 +247,13 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			inc->route = new_option->value;
 			inc->next = vpninfo->split_includes;
 			vpninfo->split_includes = inc;
+		} else if (!strcmp(buf + 7, "Split-Exclude")) {
+			struct split_include *exc = malloc(sizeof(*exc));
+			if (!exc)
+				continue;
+			exc->route = new_option->value;
+			exc->next = vpninfo->split_excludes;
+			vpninfo->split_excludes = exc;
 		}
 	}
 
@@ -274,8 +295,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpninfo->progress(vpninfo, PRG_INFO, "CSTP connected. DPD %d, Keepalive %d\n",
 			  vpninfo->ssl_times.dpd, vpninfo->ssl_times.keepalive);
 
-	BIO_set_nbio(SSL_get_rbio(vpninfo->https_ssl),1);
-	BIO_set_nbio(SSL_get_wbio(vpninfo->https_ssl),1);
+	BIO_set_nbio(SSL_get_rbio(vpninfo->https_ssl), 1);
+	BIO_set_nbio(SSL_get_wbio(vpninfo->https_ssl), 1);
 
 	fcntl(vpninfo->ssl_fd, F_SETFL, fcntl(vpninfo->ssl_fd, F_GETFL) | O_NONBLOCK);
 	if (vpninfo->select_nfds <= vpninfo->ssl_fd)
@@ -293,7 +314,7 @@ int make_cstp_connection(struct openconnect_info *vpninfo)
 {
 	int ret;
 
-	if (!vpninfo->https_ssl && (ret=openconnect_open_https(vpninfo)))
+	if (!vpninfo->https_ssl && (ret = openconnect_open_https(vpninfo)))
 		return ret;
 
 	if (vpninfo->deflate) {
@@ -327,7 +348,7 @@ static int cstp_reconnect(struct openconnect_info *vpninfo)
 	int ret;
 	int timeout;
 	int interval;
-	
+
 	timeout = vpninfo->reconnect_timeout;
 	interval = vpninfo->reconnect_interval;
 
@@ -335,7 +356,7 @@ static int cstp_reconnect(struct openconnect_info *vpninfo)
 		if (timeout <= 0)
 			return ret;
 		vpninfo->progress(vpninfo, PRG_INFO,
-				  "sleep %ds, remain timeout %ds\n",
+				  "sleep %ds, remaining timeout %ds\n",
 				  interval, timeout);
 		sleep(interval);
 		if (killed)
@@ -376,7 +397,7 @@ static int inflate_and_queue_packet(struct openconnect_info *vpninfo, int type,
 	vpninfo->inflate_adler32 = adler32(vpninfo->inflate_adler32,
 					   new->data, new->len);
 
-	if (vpninfo->inflate_adler32 != ntohl( *(uint32_t *)(buf + len - 4))) {
+	if (vpninfo->inflate_adler32 != ntohl( *(uint32_t *) (buf + len - 4) )) {
 		vpninfo->quit_reason = "Compression (inflate) adler32 failure";
 	}
 
@@ -453,7 +474,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			}
 			buf[payload_len + 8] = 0;
 			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Received server disconnect: '%s'\n", buf + 8);
+					  "Received server disconnect: %02x '%s'\n", buf[8], buf + 9);
 			vpninfo->quit_reason = "Server request";
 			return 1;
 		}
@@ -483,7 +504,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 
 	/* If SSL_write() fails we are expected to try again. With exactly
-	   the same data, at exactly the same location. So we keep the 
+	   the same data, at exactly the same location. So we keep the
 	   packet we had before.... */
 	if (vpninfo->current_ssl_pkt) {
 	handle_outgoing:
@@ -505,10 +526,9 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 					goto peer_dead;
 				return work_done;
 			default:
-				vpninfo->progress(vpninfo, PRG_ERR, "SSL_write failed: %d", ret);
-				ERR_print_errors_fp(stderr);
-				vpninfo->quit_reason = "SSL write error";
-				return 1;
+				vpninfo->progress(vpninfo, PRG_ERR, "SSL_write failed: %d\n", ret);
+				report_ssl_errors(vpninfo);
+				goto do_reconnect;
 			}
 		}
 		if (ret != vpninfo->current_ssl_pkt->len + 8) {
@@ -545,16 +565,17 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	case KA_DPD_DEAD:
 	peer_dead:
 		vpninfo->progress(vpninfo, PRG_ERR, "CSTP Dead Peer Detection detected dead peer!\n");
+	do_reconnect:
 		openconnect_close_https(vpninfo);
 
-		/* It's already deflated in the old stream. Extremely 
+		/* It's already deflated in the old stream. Extremely
 		   non-trivial to reconstitute it; just throw it away */
 		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt)
 			vpninfo->current_ssl_pkt = NULL;
 
 		if (cstp_reconnect(vpninfo)) {
 			vpninfo->progress(vpninfo, PRG_ERR, "Reconnect failed\n");
-			vpninfo->quit_reason = "SSL DPD detected dead peer; reconnect failed";
+			vpninfo->quit_reason = "CSTP reconnect failed";
 			return 1;
 		}
 		/* I think we can leave DTLS to its own devices; when we reconnect
@@ -588,6 +609,10 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpninfo->outgoing_queue = this->next;
 		vpninfo->outgoing_qlen--;
 
+		/* FIXME: Don't know how to handle IPv6 yet */
+		if (this->type != AF_INET)
+			continue;
+
 		if (vpninfo->deflate) {
 			unsigned char *adler;
 			int ret;
@@ -656,7 +681,7 @@ int cstp_bye(struct openconnect_info *vpninfo, char *reason)
 	bye_pkt = malloc(reason_len + 8);
 	if (!bye_pkt)
 		return -ENOMEM;
-	
+
 	memcpy(bye_pkt, data_hdr, 8);
 	memcpy(bye_pkt + 8, reason, reason_len);
 
