@@ -2,6 +2,7 @@
  * OpenConnect (SSL + DTLS) VPN client
  *
  * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -45,8 +46,11 @@ static void write_progress(struct openconnect_info *info, int level, const char
 static void syslog_progress(struct openconnect_info *info, int level, const char *fmt, ...);
 
 int verbose = PRG_INFO;
+int background;
+int do_passphrase_from_fsid;
 
 static struct option long_options[] = {
+	{"background", 0, 0, 'b'},
 	{"certificate", 1, 0, 'c'},
 	{"sslkey", 1, 0, 'k'},
 	{"cookie", 1, 0, 'C'},
@@ -60,8 +64,8 @@ static struct option long_options[] = {
 	{"script", 1, 0, 's'},
 	{"script-tun", 0, 0, 'S'},
 	{"syslog", 0, 0, 'l'},
-	{"tpm-key", 0, 0, 't'},
-	{"tpm-password", 1, 0, 'p'},
+	{"key-type", 1, 0, 'K'},
+	{"key-password", 1, 0, 'p'},
 	{"user", 1, 0, 'u'},
 	{"verbose", 0, 0, 'v'},
 	{"version", 0, 0, 'V'},
@@ -76,6 +80,10 @@ static struct option long_options[] = {
 	{"passwd-on-stdin", 0, 0, '5'},
 	{"no-passwd", 0, 0, '6'},
 	{"reconnect-timeout", 1, 0, '7'},
+	{"dtls-ciphers", 1, 0, '8'},
+	{"authgroup", 1, 0, '9'},
+	{"servercert", 1, 0, 0x01},
+	{"key-password-from-fsid", 0, 0, 0x02},
 	{NULL, 0, 0, 0},
 };
 
@@ -83,8 +91,10 @@ void usage(void)
 {
 	printf("Usage:  openconnect [options] <server>\n");
 	printf("Open client for Cisco AnyConnect VPN, version %s\n\n", openconnect_version);
+	printf("  -b, --background                Continue in background after startup\n");
 	printf("  -c, --certificate=CERT          Use SSL client certificate CERT\n");
 	printf("  -k, --sslkey=KEY                Use SSL private key file KEY\n");
+	printf("  -K, --key-type=TYPE             Private key type (PKCS#12 / TPM / PEM)\n");
 	printf("  -C, --cookie=COOKIE             Use WebVPN cookie COOKIE\n");
 	printf("      --cookie-on-stdin           Read cookie from standard input\n");
 	printf("  -d, --deflate                   Enable compression (default)\n");
@@ -95,23 +105,26 @@ void usage(void)
 	printf("  -l, --syslog                    Use syslog for progress messages\n");
 	printf("  -U, --setuid=USER               Drop privileges after connecting\n");
 	printf("  -m, --mtu=MTU                   Request MTU from server\n");
-	printf("  -p, --tpm-password=PASS         Set TPM SRK PIN\n");
+	printf("  -p, --key-password=PASS         Set key passphrase or TPM SRK PIN\n");
+	printf("      --key-password-from-fsid    Key passphrase is fsid of file system\n");
 	printf("  -q, --quiet                     Less output\n");
 	printf("  -Q, --queue-len=LEN             Set packet queue limit to LEN pkts\n");
 	printf("  -s, --script=SCRIPT             Use vpnc-compatible config script\n");
 	printf("  -S, --script-tun                Pass traffic to 'script' program, not tun\n");
-	printf("  -t, --tpm                       Use TPM engine for private key\n");
 	printf("  -u, --user=NAME                 Set login username\n");
 	printf("  -V, --version                   Report version number\n");
 	printf("  -v, --verbose                   More output\n");
 	printf("  -x, --xmlconfig=CONFIG          XML config file\n");
+	printf("      --authgroup=GROUP           Choose authentication login selection\n");
 	printf("      --cookieonly                Fetch webvpn cookie only; don't connect\n");
 	printf("      --printcookie               Print webvpn cookie before connecting\n");
 	printf("      --cafile=FILE               Cert file for server verification\n");
+	printf("      --dtls-ciphers=LIST         OpenSSL ciphers to support for DTLS\n");
 	printf("      --no-dtls                   Disable DTLS\n");
 	printf("      --no-passwd                 Disable password/SecurID authentication\n");
 	printf("      --passwd-on-stdin           Read password from standard input\n");
 	printf("      --reconnect-timeout         Connection retry timeout in seconds\n");
+	printf("      --servercert                Server's certificate SHA1 fingerprint\n");
 	exit(1);
 }
 
@@ -171,7 +184,7 @@ int main(int argc, char **argv)
 	else
 		vpninfo->localname = "localhost";
 
-	while ((opt = getopt_long(argc, argv, "C:c:Ddg:hi:k:lp:Q:qSs:tU:u:Vvx:",
+	while ((opt = getopt_long(argc, argv, "bC:c:Ddg:hi:k:K:lp:Q:qSs:U:u:Vvx:",
 				  long_options, NULL))) {
 		if (opt < 0)
 			break;
@@ -180,6 +193,9 @@ int main(int argc, char **argv)
 		case '0':
 			vpninfo->cafile = optarg;
 			break;
+		case 0x01:
+			vpninfo->servercert = optarg;
+			break;
 		case '1':
 			vpninfo->dtls_attempt_period = 0;
 			break;
@@ -205,6 +221,15 @@ int main(int argc, char **argv)
 		case '7':
 			vpninfo->reconnect_timeout = atoi(optarg);
 			break;
+		case '8':
+			vpninfo->dtls_ciphers = optarg;
+			break;
+		case '9':
+			vpninfo->authgroup = optarg;
+			break;
+		case 'b':
+			background = 1;
+			break;
 		case 'C':
 			vpninfo->cookie = optarg;
 			break;
@@ -214,6 +239,19 @@ int main(int argc, char **argv)
 		case 'k':
 			vpninfo->sslkey = optarg;
 			break;
+		case 'K':
+			if (!strcasecmp(optarg, "PKCS#12") ||
+			    !strcasecmp(optarg, "PKCS12")) {
+				vpninfo->cert_type = CERT_TYPE_PKCS12;
+			} else if (!strcasecmp(optarg, "TPM")) {
+				vpninfo->cert_type = CERT_TYPE_TPM;
+			} else if (!strcasecmp(optarg, "PEM")) {
+				vpninfo->cert_type = CERT_TYPE_PEM;
+			} else {
+				fprintf(stderr, "Unknown certificate type '%s'\n",
+					optarg);
+				usage();
+			}
 		case 'd':
 			vpninfo->deflate = 1;
 			break;
@@ -240,7 +278,7 @@ int main(int argc, char **argv)
 			}
 			break;
 		case 'p':
-			vpninfo->tpmpass = optarg;
+			vpninfo->cert_password = optarg;
 			break;
 		case 's':
 			vpninfo->vpnc_script = optarg;
@@ -248,9 +286,6 @@ int main(int argc, char **argv)
 		case 'S':
 			vpninfo->script_tun = 1;
 			break;
-		case 't':
-			vpninfo->tpm = 1;
-			break;
 		case 'u':
 			vpninfo->username = optarg;
 			break;
@@ -288,6 +323,9 @@ int main(int argc, char **argv)
 			vpninfo->xmlconfig = optarg;
 			vpninfo->write_new_config = write_new_config;
 			break;
+		case 0x02:
+			do_passphrase_from_fsid = 1;
+			break;
 		default:
 			usage();
 		}
@@ -307,6 +345,9 @@ int main(int argc, char **argv)
 		vpninfo->progress = write_progress;
 	}
 
+	if (vpninfo->sslkey && do_passphrase_from_fsid)
+		passphrase_from_fsid(vpninfo);
+
 	if (config_lookup_host(vpninfo, argv[optind]))
 		exit(1);
 
@@ -317,7 +358,7 @@ int main(int argc, char **argv)
 	set_openssl_ui();
 #endif
 
-	if (!vpninfo->cookie && openconnect_obtain_cookie(vpninfo) != 1) {
+	if (!vpninfo->cookie && openconnect_obtain_cookie(vpninfo)) {
 		fprintf(stderr, "Failed to obtain WebVPN cookie\n");
 		exit(1);
 	}
@@ -332,7 +373,7 @@ int main(int argc, char **argv)
 		fprintf(stderr, "Creating SSL connection failed\n");
 		exit(1);
 	}
-	
+
 	if (setup_tun(vpninfo)) {
 		fprintf(stderr, "Set up tun device failed\n");
 		exit(1);
@@ -344,15 +385,26 @@ int main(int argc, char **argv)
 			exit(1);
 		}
 	}
-		
+
 	if (vpninfo->dtls_attempt_period && setup_dtls(vpninfo))
 		fprintf(stderr, "Set up DTLS failed; using SSL instead\n");
 
 	vpninfo->progress(vpninfo, PRG_INFO,
 			  "Connected %s as %s, using %s\n", vpninfo->ifname,
 			  vpninfo->vpn_addr,
-			  (vpninfo->dtls_fd==-1)?(vpninfo->deflate?"SSL + deflate":"SSL"):"DTLS");
+			  (vpninfo->dtls_fd == -1) ?
+			      (vpninfo->deflate ? "SSL + deflate" : "SSL")
+			      : "DTLS");
 
+	if (background) {
+		int pid;
+		if ((pid = fork ())) {
+			vpninfo->progress(vpninfo, PRG_INFO,
+					  "Continuing in background; pid %d\n",
+					  pid);
+			exit (1);
+		}
+	}
 	vpn_mainloop(vpninfo);
 	exit(1);
 }
@@ -363,19 +415,19 @@ static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buf
 
 	config_fd = open(vpninfo->xmlconfig, O_WRONLY|O_TRUNC|O_CREAT, 0644);
 	if (!config_fd) {
-		fprintf(stderr, "Failed to open %s for write: %s\n", 
+		fprintf(stderr, "Failed to open %s for write: %s\n",
 			vpninfo->xmlconfig, strerror(errno));
 		return -errno;
 	}
 
 	/* FIXME: We should actually write to a new tempfile, then rename */
-	write(config_fd, buf, buflen);	
+	write(config_fd, buf, buflen);
 	return 0;
 }
 
 void write_progress(struct openconnect_info *info, int level, const char *fmt, ...)
 {
-	FILE *outf = level?stdout:stderr;
+	FILE *outf = level ? stdout : stderr;
 	va_list args;
 
 	if (verbose >= level) {
@@ -388,7 +440,7 @@ void write_progress(struct openconnect_info *info, int level, const char *fmt, .
 void syslog_progress(struct openconnect_info *info, int level,
 		     const char *fmt, ...)
 {
-	int priority = level?LOG_INFO:LOG_NOTICE;
+	int priority = level ? LOG_INFO : LOG_NOTICE;
 	va_list args;
 
 	if (verbose >= level) {
