@@ -29,6 +29,9 @@
 #include <signal.h>
 #include <arpa/inet.h>
 #include <unistd.h>
+#include <string.h>
+
+#include <openssl/ssl.h>
 
 #include "openconnect.h"
 
@@ -67,7 +70,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 	struct sigaction sa;
 	memset(&sa, 0, sizeof(sa));
 	sa.sa_handler = handle_sigint;
-	
+
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGHUP, &sa, NULL);
 
@@ -77,6 +80,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		struct timeval tv;
 		fd_set rfds, wfds, efds;
 
+#ifdef SSL_F_DTLS1_CONNECT
 		if (vpninfo->new_dtls_ssl)
 			dtls_try_handshake(vpninfo);
 
@@ -87,7 +91,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		}
 		if (vpninfo->dtls_ssl)
 			did_work += dtls_mainloop(vpninfo, &timeout);
-
+#endif
 		if (vpninfo->quit_reason)
 			break;
 
@@ -114,7 +118,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		if (did_work)
 			continue;
 
-		vpninfo->progress(vpninfo, PRG_TRACE, 
+		vpninfo->progress(vpninfo, PRG_TRACE,
 				  "Did no work; sleeping for %d ms...\n", timeout);
 		memcpy(&rfds, &vpninfo->select_rfds, sizeof(rfds));
 		memcpy(&wfds, &vpninfo->select_wfds, sizeof(wfds));
@@ -123,10 +127,13 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		tv.tv_sec = timeout / 1000;
 		tv.tv_usec = (timeout % 1000) * 1000;
 		if (select(vpninfo->select_nfds, &rfds, &wfds, &efds, &tv) >= 0
-		    && FD_ISSET(vpninfo->ssl_fd, &efds)) {
-			/* OpenSSL doesn't seem to cope properly with this... */
-			vpninfo->progress(vpninfo, PRG_ERR, "Server closed connection!\n");
-			exit(1);
+		    && FD_ISSET(vpninfo->ssl_fd, &rfds)) {
+			char b;
+			if (recv(vpninfo->ssl_fd, &b, 1, MSG_PEEK) == 0) {
+				/* Zero indicates EOF, which OpenSSL won't handle on read() */
+				vpninfo->progress(vpninfo, PRG_ERR, "Server closed connection!\n");
+				exit(1);
+			}
 		}
 	}
 
