@@ -5,7 +5,7 @@
  *
  * Authors: Jussi Kukkonen <jku@linux.intel.com>
  *          David Woodhouse <dwmw2@infradead.org>
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * version 2.1, as published by the Free Software Foundation.
@@ -64,9 +64,15 @@ enum certificate_response{
 	CERT_ACCEPTED,
 };
 
+struct gconf_key {
+	char *key;
+	char *value;
+	struct gconf_key *next;
+};
+
 typedef struct auth_ui_data {
 	struct openconnect_info *vpninfo;
-	char *firsthost;
+	struct gconf_key *success_keys;
 	GtkWidget *dialog;
 	GtkWidget *combo;
 	GtkWidget *connect_button;
@@ -79,10 +85,11 @@ typedef struct auth_ui_data {
 
 	int retval;
 	int cookie_retval;
-	
+
 	gboolean cancelled; /* fully cancel the whole challenge-response series */
 	gboolean getting_cookie;
-	
+
+	int form_grabbed;
 	GQueue *form_entries; /* modified from worker thread */
 	GMutex *form_mutex;
 
@@ -99,7 +106,7 @@ typedef struct auth_ui_data {
 enum {
 	AUTH_DIALOG_RESPONSE_LOGIN = 1,
 	AUTH_DIALOG_RESPONSE_CANCEL,
-}auth_dialog_response;
+} auth_dialog_response;
 
 
 
@@ -135,20 +142,36 @@ static void ssl_box_add_error(auth_ui_data *ui_data, const char *msg)
 	gtk_box_pack_start(GTK_BOX(hbox), text, FALSE, FALSE, 0);
 }
 
+static void ssl_box_add_info(auth_ui_data *ui_data, const char *msg)
+{
+	GtkWidget *text;
+	int width;
+
+	text = gtk_label_new(msg);
+	gtk_label_set_line_wrap(GTK_LABEL(text), TRUE);
+	gtk_window_get_size(GTK_WINDOW(ui_data->dialog), &width, NULL);
+	/* FIXME: this is not very nice -- can't make the window thinner after this */
+	gtk_widget_set_size_request(text, width - 40, -1);
+	gtk_box_pack_start(GTK_BOX(ui_data->ssl_box), text, FALSE, FALSE, 0);
+}
+
 static void ssl_box_clear(auth_ui_data *ui_data)
 {
 	gtk_widget_hide(ui_data->no_form_label);
 	gtk_widget_hide(ui_data->getting_form_label);
-	gtk_container_foreach(GTK_CONTAINER(ui_data->ssl_box), 
+	gtk_container_foreach(GTK_CONTAINER(ui_data->ssl_box),
 			      container_child_remove, ui_data->ssl_box);
 	gtk_widget_set_sensitive (ui_data->login_button, FALSE);
 	gtk_widget_set_sensitive (ui_data->cancel_button, FALSE);
 }
 
 typedef struct ui_fragment_data {
+	GtkWidget *widget;
 	auth_ui_data *ui_data;
 	UI_STRING *uis;
+	struct oc_form_opt *opt;
 	char *entry_text;
+	int grab_focus;
 } ui_fragment_data;
 
 static void entry_activate_cb(GtkWidget *widget, auth_ui_data *ui_data)
@@ -156,10 +179,59 @@ static void entry_activate_cb(GtkWidget *widget, auth_ui_data *ui_data)
 	gtk_dialog_response(GTK_DIALOG(ui_data->dialog), AUTH_DIALOG_RESPONSE_LOGIN);
 }
 
+static void do_check_visibility(ui_fragment_data *data, gboolean *visible)
+{
+	int min_len;
+
+	if (!data->uis)
+		return;
+
+	min_len = UI_get_result_minsize(data->uis);
+
+	if (min_len && (!data->entry_text || strlen(data->entry_text) < min_len))
+		*visible = FALSE;
+}
+
+static void evaluate_login_visibility(auth_ui_data *ui_data)
+{
+	gboolean visible = TRUE;
+	g_queue_foreach(ui_data->form_entries, (GFunc)do_check_visibility,
+			&visible);
+
+	gtk_widget_set_sensitive (ui_data->login_button, visible);
+}
+
 static void entry_changed(GtkEntry *entry, ui_fragment_data *data)
 {
 	g_free (data->entry_text);
 	data->entry_text = g_strdup(gtk_entry_get_text(entry));
+	evaluate_login_visibility(data->ui_data);
+}
+
+static void do_override_label(ui_fragment_data *data, struct oc_choice *choice)
+{
+	const char *new_label = data->opt->label;
+
+	if (!data->widget)
+		return;
+
+	if (choice->override_name && !strcmp(choice->override_name, data->opt->name))
+		    new_label = choice->override_label;
+
+	gtk_label_set_text(GTK_LABEL(data->widget), new_label);
+
+}
+static void combo_changed(GtkComboBox *combo, ui_fragment_data *data)
+{
+	struct oc_form_opt_select *sopt = (void *)data->opt;
+	int entry = gtk_combo_box_get_active(combo);
+	if (entry < 0)
+		return;
+
+	data->entry_text = sopt->choices[entry].name;
+
+	g_queue_foreach(data->ui_data->form_entries, (GFunc)do_override_label,
+			&sopt->choices[entry]);
 }
 
 static gboolean ui_write_error (ui_fragment_data *data)
@@ -173,15 +245,7 @@ static gboolean ui_write_error (ui_fragment_data *data)
 
 static gboolean ui_write_info (ui_fragment_data *data)
 {
-	GtkWidget *text;
-	int width;
-
-	text = gtk_label_new(UI_get0_output_string(data->uis));
-	gtk_label_set_line_wrap(GTK_LABEL(text), TRUE);
-	gtk_window_get_size(GTK_WINDOW(ui_data->dialog), &width, NULL);
-	/* FIXME: this is not very nice -- can't make the window thinner after this */
-	gtk_widget_set_size_request(text, width - 40, -1);
-	gtk_box_pack_start(GTK_BOX(data->ui_data->ssl_box), text, FALSE, FALSE, 0);
+	ssl_box_add_info(data->ui_data, UI_get0_output_string(data->uis));
 
 	g_slice_free (ui_fragment_data, data);
 
@@ -191,19 +255,34 @@ static gboolean ui_write_info (ui_fragment_data *data)
 static gboolean ui_write_prompt (ui_fragment_data *data)
 {
 	GtkWidget *hbox, *text, *entry;
+	int visible;
+	const char *label;
+
+	if (data->uis) {
+		label = UI_get0_output_string(data->uis);
+		visible = UI_get_input_flags(data->uis) & UI_INPUT_FLAG_ECHO;
+	} else {
+		label = data->opt->label;
+		visible = (data->opt->type == OC_FORM_OPT_TEXT);
+	}
 
 	hbox = gtk_hbox_new(FALSE, 0);
 	gtk_box_pack_start(GTK_BOX(data->ui_data->ssl_box), hbox, FALSE, FALSE, 0);
 
-	text = gtk_label_new(UI_get0_output_string(data->uis));
+	text = gtk_label_new(label);
 	gtk_box_pack_start(GTK_BOX(hbox), text, FALSE, FALSE, 0);
+	data->widget = text;
 
 	entry = gtk_entry_new();
 	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
-	if (!(UI_get_input_flags(data->uis) & UI_INPUT_FLAG_ECHO))
+	if (!visible)
 		gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
-	if (g_queue_peek_tail(ui_data->form_entries) == data)
+	if (data->entry_text)
+		gtk_entry_set_text(GTK_ENTRY(entry), data->entry_text);
+	if (!data->entry_text && !data->ui_data->form_grabbed) {
+		data->ui_data->form_grabbed = 1;
 		gtk_widget_grab_focus (entry);
+	}
 	g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(entry_changed), data);
 	g_signal_connect(G_OBJECT(entry), "activate", G_CALLBACK(entry_activate_cb), ui_data);
 
@@ -212,14 +291,53 @@ static gboolean ui_write_prompt (ui_fragment_data *data)
 	return FALSE;
 }
 
+static gboolean ui_add_select (ui_fragment_data *data)
+{
+	GtkWidget *hbox, *text, *combo;
+	struct oc_form_opt_select *sopt = (void *)data->opt;
+	int i;
+
+	hbox = gtk_hbox_new(FALSE, 0);
+	gtk_box_pack_start(GTK_BOX(data->ui_data->ssl_box), hbox, FALSE, FALSE, 0);
+
+	text = gtk_label_new(data->opt->label);
+	gtk_box_pack_start(GTK_BOX(hbox), text, FALSE, FALSE, 0);
+
+	combo = gtk_combo_box_new_text();
+	gtk_box_pack_end(GTK_BOX(hbox), combo, FALSE, FALSE, 0);
+	for (i = 0; i < sopt->nr_choices; i++) {
+		gtk_combo_box_append_text(GTK_COMBO_BOX(combo), sopt->choices[i].label);
+		if (data->entry_text && 
+		    !strcmp(data->entry_text, sopt->choices[i].name)) {
+			gtk_combo_box_set_active(GTK_COMBO_BOX(combo), i);
+			g_free(data->entry_text);
+			data->entry_text = sopt->choices[i].name;
+		}
+	}
+	if (gtk_combo_box_get_active(GTK_COMBO_BOX(combo)) < 0) {
+		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0); 
+		data->entry_text = sopt->choices[0].name;
+	}
+
+	if (g_queue_peek_tail(ui_data->form_entries) == data)
+		gtk_widget_grab_focus (combo);
+	g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(combo_changed), data);
+	/* Hook up the 'show' signal to ensure that we override prompts on 
+	   UI elements which may be coming later. */
+	g_signal_connect(G_OBJECT(combo), "show", G_CALLBACK(combo_changed), data);
+
+	/* data is freed in ui_flush in worker thread */
+
+	return FALSE;
+}
+
 static gboolean ui_show (auth_ui_data *ui_data)
 {
 	gtk_widget_hide (ui_data->getting_form_label);
 	gtk_widget_show_all (ui_data->ssl_box);
 	gtk_widget_set_sensitive (ui_data->cancel_button, TRUE);
 	g_mutex_lock (ui_data->form_mutex);
-	if (!g_queue_is_empty(ui_data->form_entries))
-		gtk_widget_set_sensitive (ui_data->login_button, TRUE);
+	evaluate_login_visibility(ui_data);
 	ui_data->form_shown = TRUE;
 	g_cond_signal (ui_data->form_shown_changed);
 	g_mutex_unlock (ui_data->form_mutex);
@@ -296,25 +414,25 @@ static int ui_flush(UI* ui)
 	ui_data->form_shown = FALSE;
 
 	if (!ui_data->cancelled) {
-		/* wait for form submission or cancel */ 
+		/* wait for form submission or cancel */
 		while (!ui_data->form_retval) {
 			g_cond_wait(ui_data->form_retval_changed, ui_data->form_mutex);
 		}
 		response = GPOINTER_TO_INT (ui_data->form_retval);
 		ui_data->form_retval = NULL;
-	} else 
+	} else
 		response = AUTH_DIALOG_RESPONSE_CANCEL;
 
 	/* set entry results and free temporary data structures */
 	while (!g_queue_is_empty (ui_data->form_entries)) {
 		ui_fragment_data *data;
-		data = g_queue_pop_tail (ui_data->form_entries); 
+		data = g_queue_pop_tail (ui_data->form_entries);
 		if (data->entry_text) {
 			UI_set_result(ui, data->uis, data->entry_text);
 		}
 		g_slice_free (ui_fragment_data, data);
 	}
-
+	ui_data->form_grabbed = 0;
 	g_mutex_unlock(ui_data->form_mutex);
 
 	/* -1 = cancel,
@@ -342,6 +460,142 @@ int init_openssl_ui(void)
 	return 0;
 }
 
+void remember_gconf_key(auth_ui_data *ui_data, char *key, char *value)
+{
+	struct gconf_key *k = g_malloc(sizeof(*k));
+
+	if (!k)
+		return;
+
+	k->next = ui_data->success_keys;
+	k->key = key;
+	k->value = value;
+
+	ui_data->success_keys = k;
+	while (k->next) {
+		if (!strcmp(k->next->key, key)) {
+			struct gconf_key *old = k->next;
+			k->next = old->next;
+			g_free(old->key);
+			g_free(old->value);
+			g_free(old);
+			break;
+		}
+		k = k->next;
+	}
+}
+
+char *find_form_answer(struct oc_auth_form *form, struct oc_form_opt *opt)
+{
+	char *key, *result;
+	key = g_strdup_printf("%s/vpn/form:%s:%s", config_path,
+			      form->auth_id, opt->name);
+	result = gconf_client_get_string(gcl, key, NULL);
+	g_free(key);
+	return result;
+}
+
+/* This part for processing forms from openconnect directly, rather than
+   through the SSL UI abstraction (which doesn't allow 'select' options) */
+
+static gboolean ui_form (struct oc_auth_form *form)
+{
+	struct oc_form_opt *opt;
+
+	ssl_box_clear(ui_data);
+
+	if (form->banner)
+		ssl_box_add_info(ui_data, form->banner);
+	if (form->error)
+		ssl_box_add_error(ui_data, form->error);
+	if (form->message)
+		ssl_box_add_info(ui_data, form->message);
+
+	for (opt = form->opts; opt; opt = opt->next) {
+		ui_fragment_data *data;
+
+		if (opt->type == OC_FORM_OPT_HIDDEN)
+			continue;
+
+		data = g_slice_new0 (ui_fragment_data);
+		data->ui_data = ui_data;
+		data->opt = opt;
+		
+		if (opt->type == OC_FORM_OPT_PASSWORD ||
+		    opt->type == OC_FORM_OPT_TEXT) {
+			g_mutex_lock (ui_data->form_mutex);
+			g_queue_push_head(ui_data->form_entries, data);
+			g_mutex_unlock (ui_data->form_mutex);
+			if (opt->type != OC_FORM_OPT_PASSWORD)
+				data->entry_text = find_form_answer(form, opt);
+
+			ui_write_prompt(data);
+		} else if (opt->type == OC_FORM_OPT_SELECT) {
+			g_mutex_lock (ui_data->form_mutex);
+			g_queue_push_head(ui_data->form_entries, data);
+			g_mutex_unlock (ui_data->form_mutex);
+			data->entry_text = find_form_answer(form, opt);
+
+			ui_add_select(data);
+		} else
+			g_slice_free (ui_fragment_data, data);
+	}
+	
+	return ui_show(ui_data);
+}
+
+int nm_process_auth_form (struct openconnect_info *vpninfo,
+			  struct oc_auth_form *form)
+{
+	int response;
+
+	g_idle_add((GSourceFunc)ui_form, form);
+
+	g_mutex_lock(ui_data->form_mutex);
+	/* wait for ui to show */
+	while (!ui_data->form_shown) {
+		g_cond_wait(ui_data->form_shown_changed, ui_data->form_mutex);
+	}
+	ui_data->form_shown = FALSE;
+
+	if (!ui_data->cancelled) {
+		/* wait for form submission or cancel */
+		while (!ui_data->form_retval) {
+			g_cond_wait(ui_data->form_retval_changed, ui_data->form_mutex);
+		}
+		response = GPOINTER_TO_INT (ui_data->form_retval);
+		ui_data->form_retval = NULL;
+	} else
+		response = AUTH_DIALOG_RESPONSE_CANCEL;
+
+	if (response == AUTH_DIALOG_RESPONSE_LOGIN) {
+		/* set entry results and free temporary data structures */
+		while (!g_queue_is_empty (ui_data->form_entries)) {
+			ui_fragment_data *data;
+			data = g_queue_pop_tail (ui_data->form_entries);
+			if (data->entry_text) {
+				data->opt->value = data->entry_text;
+
+				if (data->opt->type == OC_FORM_OPT_TEXT ||
+				    data->opt->type == OC_FORM_OPT_SELECT) {
+					char *keyname;
+					keyname = g_strdup_printf("form:%s:%s", form->auth_id, data->opt->name);
+					remember_gconf_key(ui_data, keyname, strdup(data->entry_text));
+				}
+			}
+			g_slice_free (ui_fragment_data, data);
+		}
+	}
+
+
+	g_mutex_unlock(ui_data->form_mutex);
+	
+	/* -1 = cancel,
+	 *  0 = failure,
+	 *  1 = success */
+	return (response == AUTH_DIALOG_RESPONSE_LOGIN ? 0 : 1);
+
+}
 
 static char* get_title(const char *vpn_name)
 {
@@ -411,7 +665,7 @@ static gboolean user_validate_cert(cert_data *data)
 	g_mutex_lock (ui_data->form_mutex);
 	if (result == GTK_RESPONSE_OK)
 		data->ui_data->cert_response = CERT_ACCEPTED;
-	else 
+	else
 		data->ui_data->cert_response = CERT_DENIED;
 	g_cond_signal (ui_data->cert_response_changed);
 	g_mutex_unlock (ui_data->form_mutex);
@@ -423,18 +677,15 @@ static gboolean user_validate_cert(cert_data *data)
 static int validate_peer_cert(struct openconnect_info *vpninfo,
 			      X509 *peer_cert)
 {
-	ASN1_STRING *signature = peer_cert->signature;
+	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
 	char *certs_data;
 	char *key;
-	BIO *bp = BIO_new(BIO_s_mem());
-	BUF_MEM *sig;
-	char zero = 0;
 	int ret = 0;
 	cert_data *data;
 
-	i2a_ASN1_STRING(bp, signature, V_ASN1_OCTET_STRING);
-	BIO_write(bp, &zero, 1);
-	BIO_get_mem_ptr(bp, &sig);
+	ret = get_cert_fingerprint(peer_cert, fingerprint);
+	if (ret)
+		return ret;
 
 	key = g_strdup_printf("%s/vpn/%s", config_path, "certsigs");
 	certs_data = gconf_client_get_string(gcl, key, NULL);
@@ -443,7 +694,7 @@ static int validate_peer_cert(struct openconnect_info *vpninfo,
 		char **this = certs;
 
 		while (*this) {
-			if (!strcmp(*this, sig->data)) {
+			if (!strcmp(*this, fingerprint)) {
 				g_strfreev(certs);
 				goto out;
 			}
@@ -458,20 +709,20 @@ static int validate_peer_cert(struct openconnect_info *vpninfo,
 
 	g_mutex_lock(ui_data->form_mutex);
 
-	ui_data->cert_response = CERT_USER_NOT_READY; 
+	ui_data->cert_response = CERT_USER_NOT_READY;
 	g_idle_add((GSourceFunc)user_validate_cert, data);
 
-	/* wait for user to accept or cancel */ 
+	/* wait for user to accept or cancel */
 	while (ui_data->cert_response == CERT_USER_NOT_READY) {
 		g_cond_wait(ui_data->cert_response_changed, ui_data->form_mutex);
 	}
 	if (ui_data->cert_response == CERT_ACCEPTED) {
 		if (certs_data) {
-			char *new = g_strdup_printf("%s\t%s", certs_data, sig->data);
+			char *new = g_strdup_printf("%s\t%s", certs_data, fingerprint);
 			gconf_client_set_string(gcl, key, new, NULL);
 			g_free(new);
 		} else {
-			gconf_client_set_string(gcl, key, sig->data, NULL);
+			gconf_client_set_string(gcl, key, fingerprint, NULL);
 		}
 		ret = 0;
 	} else {
@@ -482,9 +733,7 @@ static int validate_peer_cert(struct openconnect_info *vpninfo,
 	g_slice_free(cert_data, data);
 
  out:
-	if (certs_data)
-		g_free(certs_data);
-	BIO_free(bp);
+	g_free(certs_data);
 	g_free(key);
 	return ret;
 }
@@ -607,7 +856,7 @@ static int parse_xmlconfig(char *xmlconfig)
 						list_end = &newhost->next;
 
 						if (!strcasecmp(newhost->hostaddress, vpnhosts->hostaddress) &&
-						    !strcasecmp(newhost->usergroup?:"", vpnhosts->usergroup?:"")) {
+						    !strcasecmp(newhost->usergroup ?: "", vpnhosts->usergroup ?: "")) {
 							/* Remove originally configured host if it's in the list */
 							struct vpnhost *tmp = vpnhosts->next;
 							free(vpnhosts);
@@ -631,6 +880,7 @@ static int get_config(char *vpn_uuid, struct openconnect_info *vpninfo)
 	char *xmlconfig;
 	char *hostname;
 	char *group;
+	char *pem_passphrase_fsid;
 
 	gcl = gconf_client_get_default();
 	config_path = get_config_path(gcl, vpn_uuid);
@@ -659,9 +909,9 @@ static int get_config(char *vpn_uuid, struct openconnect_info *vpninfo)
 	vpnhosts->hostaddress = hostname;
 	vpnhosts->next = NULL;
 
-if (0){
+if (0) {
 /* DEBUG add another copy of gateway to host list */
-	 vpnhost *tmphost;
+	vpnhost *tmphost;
 	tmphost = malloc(sizeof(tmphost));
 	if (!tmphost)
 		return -ENOMEM;
@@ -704,7 +954,7 @@ if (0){
 		return 0;
 	}
 	if (!strcmp(authtype, NM_OPENCONNECT_AUTHTYPE_CERT_TPM))
-		vpninfo->tpm = 1;
+		vpninfo->cert_type = CERT_TYPE_TPM;
 	else if (strcmp(authtype, NM_OPENCONNECT_AUTHTYPE_CERT)) {
 		fprintf(stderr, "Unknown authentication type '%s'\n", authtype);
 		return -EINVAL;
@@ -721,10 +971,15 @@ if (0){
 	if (!vpninfo->sslkey)
 		vpninfo->sslkey = vpninfo->cert;
 
+	pem_passphrase_fsid = get_gconf_setting(gcl, config_path, "pem_passphrase_fsid");
+	if (pem_passphrase_fsid && !strcmp(pem_passphrase_fsid, "yes"))
+		passphrase_from_fsid(vpninfo);
+	g_free(pem_passphrase_fsid);
+				    
 	return 0;
 }
 
-static void populate_vpnhost_combo(auth_ui_data *ui_data) 
+static void populate_vpnhost_combo(auth_ui_data *ui_data)
 {
 	struct vpnhost *host;
 	int i = 0;
@@ -754,7 +1009,7 @@ static void autocon_toggled(GtkWidget *widget)
 	int enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
 	char *key = g_strdup_printf("%s/vpn/autoconnect", config_path);
 
-	gconf_client_set_string(gcl, key, enabled?"yes":"no", NULL);
+	gconf_client_set_string(gcl, key, enabled ? "yes" : "no", NULL);
 }
 
 static void scroll_log(GtkTextBuffer *log, GtkTextView *view)
@@ -801,13 +1056,22 @@ void write_progress(struct openconnect_info *info, int level, const char *fmt, .
 		g_idle_add((GSourceFunc)write_progress_real, g_strdup(msg));
 	}
 
-	if (level <= PRG_ERR){
+	if (level <= PRG_ERR) {
 		last_message = msg;
 		return;
-	} 
+	}
 	g_free(msg);
 }
 
+static void print_peer_cert(struct openconnect_info *vpninfo)
+{
+	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	X509 *cert = SSL_get_peer_certificate(vpninfo->https_ssl);
+
+	if (cert && !get_cert_fingerprint(cert, fingerprint))
+		printf("gwcert\n%s\n", fingerprint);
+}
+
 static gboolean cookie_obtained(auth_ui_data *ui_data)
 {
 	ui_data->getting_cookie = FALSE;
@@ -815,6 +1079,14 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 
 	if (ui_data->cancelled) {
 		/* user has chosen a new host, start from beginning */
+		while (ui_data->success_keys) {
+			struct gconf_key *k = ui_data->success_keys;
+			
+			ui_data->success_keys = k->next;
+			g_free(k->key);
+			g_free(k->value);
+			g_free(k);
+		}			
 		connect_host(ui_data);
 		return FALSE;
 	}
@@ -827,15 +1099,24 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 			gtk_widget_set_sensitive(ui_data->cancel_button, TRUE);
 		}
 		ui_data->retval = 1;
-	} else if (ui_data->cookie_retval == 1) {
+	} else if (!ui_data->cookie_retval) {
 		/* got cookie */
-		char *key = g_strdup_printf("%s/vpn/lasthost", config_path);
-		/* We don't use vpninfo->hostname because it might have been redirected */
-		gconf_client_set_string(gcl, key, ui_data->firsthost, NULL);
-		g_free(key);
-		
+		while (ui_data->success_keys) {
+			struct gconf_key *k = ui_data->success_keys;
+			char *key = g_strdup_printf("%s/vpn/%s", config_path, k->key);
+
+			gconf_client_set_string(gcl, key, k->value, NULL);
+			g_free(key);
+
+			ui_data->success_keys = k->next;
+			g_free(k->key);
+			g_free(k->value);
+			g_free(k);
+		}
+
 		printf("%s\n%s\n", NM_OPENCONNECT_KEY_GATEWAY, ui_data->vpninfo->hostname);
 		printf("%s\n%s\n", NM_OPENCONNECT_KEY_COOKIE, ui_data->vpninfo->cookie);
+		print_peer_cert(ui_data->vpninfo);
 		memset((void *)ui_data->vpninfo->cookie, 0, strlen(ui_data->vpninfo->cookie));
 		printf("\n\n");
 		fflush(stdout);
@@ -843,11 +1124,20 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 
 		gtk_main_quit();
 	} else {
-		/* no cookie */
+		/* no cookie; user cancellation */
 		gtk_widget_show (ui_data->no_form_label);
 		ui_data->retval = 1;
 	}
 
+	while (ui_data->success_keys) {
+		struct gconf_key *k = ui_data->success_keys;
+
+		ui_data->success_keys = k->next;
+		g_free(k->key);
+		g_free(k->value);
+		g_free(k);
+	}			
+
 	return FALSE;
 }
 
@@ -880,12 +1170,12 @@ static void connect_host(auth_ui_data *ui_data)
 	ssl_box_clear(ui_data);
 	gtk_widget_show(ui_data->getting_form_label);
 
-	/* reset ssl context. 
+	/* reset ssl context.
 	 * TODO: this is probably not the way to go... */
 	if (ui_data->vpninfo->https_ssl) {
 		openconnect_close_https(ui_data->vpninfo);
 	}
-	if (ui_data->vpninfo->https_ctx){
+	if (ui_data->vpninfo->https_ctx) {
 		SSL_CTX_free(ui_data->vpninfo->https_ctx);
 		ui_data->vpninfo->https_ctx = NULL;
 	}
@@ -898,7 +1188,8 @@ static void connect_host(auth_ui_data *ui_data)
 	ui_data->vpninfo->hostname = g_strdup(host->hostaddress);
 	if (host->usergroup)
 		ui_data->vpninfo->urlpath = g_strdup(host->usergroup);
-	ui_data->firsthost = g_strdup(host->hostname);
+
+	remember_gconf_key(ui_data, g_strdup("lasthost"), g_strdup(host->hostname));
 
 	thread = g_thread_create((GThreadFunc)obtain_cookie, ui_data,
 				 FALSE, NULL);
@@ -915,7 +1206,7 @@ static void queue_connect_host(auth_ui_data *ui_data)
 		connect_host(ui_data);
 	} else if (!ui_data->cancelled) {
 		/* set state to cancelled. Current challenge-response-
-		 * conversation will not be shown to user, and cookie_obtained() 
+		 * conversation will not be shown to user, and cookie_obtained()
 		 * will start a new one conversation */
 		ui_data->cancelled = TRUE;
 		gtk_dialog_response(GTK_DIALOG(ui_data->dialog), AUTH_DIALOG_RESPONSE_CANCEL);
@@ -928,7 +1219,7 @@ static void dialog_response (GtkDialog *dialog, int response, auth_ui_data *ui_d
 	case AUTH_DIALOG_RESPONSE_CANCEL:
 	case AUTH_DIALOG_RESPONSE_LOGIN:
 		ssl_box_clear(ui_data);
-		if(ui_data->getting_cookie)
+		if (ui_data->getting_cookie)
 			gtk_widget_show (ui_data->getting_form_label);
 		g_mutex_lock (ui_data->form_mutex);
 		ui_data->form_retval = GINT_TO_POINTER(response);
@@ -962,7 +1253,7 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	gtk_window_set_default_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION);
 
 	title = get_title(ui_data->vpninfo->vpn_name);
-	ui_data->dialog = gtk_dialog_new_with_buttons(title, NULL, GTK_DIALOG_MODAL, 
+	ui_data->dialog = gtk_dialog_new_with_buttons(title, NULL, GTK_DIALOG_MODAL,
 						      GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
 						      NULL);
 	g_signal_connect (ui_data->dialog, "response", G_CALLBACK(dialog_response), ui_data);
@@ -987,7 +1278,7 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	ui_data->combo = gtk_combo_box_new_text();
 	populate_vpnhost_combo(ui_data);
 	gtk_box_pack_start(GTK_BOX(hbox), ui_data->combo, TRUE, TRUE, 0);
-	g_signal_connect_swapped(ui_data->combo, "changed", 
+	g_signal_connect_swapped(ui_data->combo, "changed",
 	                         G_CALLBACK(queue_connect_host), ui_data);
 	gtk_widget_show(ui_data->combo);
 
@@ -1009,7 +1300,7 @@ static void build_main_dialog(auth_ui_data *ui_data)
 
 	frame = gtk_frame_new(NULL);
 	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);
-	gtk_widget_set_size_request(frame, -1, 175);
+	gtk_widget_set_size_request(frame, -1, -1);
 	gtk_widget_show(frame);
 
 	frame_box = gtk_vbox_new(FALSE, 4);
@@ -1077,7 +1368,7 @@ static auth_ui_data *init_ui_data (char *vpn_name)
 {
 	auth_ui_data *ui_data;
 
-	ui_data = g_slice_new0(auth_ui_data); 
+	ui_data = g_slice_new0(auth_ui_data);
 	ui_data->retval = 1;
 
 	ui_data->form_entries = g_queue_new();
@@ -1094,6 +1385,7 @@ static auth_ui_data *init_ui_data (char *vpn_name)
 	ui_data->vpninfo->progress = write_progress;
 	ui_data->vpninfo->validate_peer_cert = validate_peer_cert;
 	ui_data->vpninfo->vpn_name = vpn_name;
+	ui_data->vpninfo->process_auth_form = nm_process_auth_form;
 
 	return ui_data;
 }
