@@ -2,6 +2,7 @@
  * OpenConnect (SSL + DTLS) VPN client
  *
  * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -34,6 +35,55 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+/****************************************************************************/
+
+/* Authentication form processing */
+
+#define OC_FORM_OPT_TEXT	1
+#define OC_FORM_OPT_PASSWORD	2
+#define OC_FORM_OPT_SELECT	3
+#define OC_FORM_OPT_HIDDEN	4
+
+/* char * fields are static (owned by XML parser) and don't need to be
+   freed by the form handling code -- except for value, which for TEXT
+   and PASSWORD options is allocated by process_form() when
+   interacting with the user and must be freed. */
+struct oc_form_opt {
+	struct oc_form_opt *next;
+	int type;
+	char *name;
+	char *label;
+	char *value;
+};
+
+/* All fields are static, owned by the XML parser */
+struct oc_choice {
+	char *name;
+	char *label;
+	char *auth_type;
+	char *override_name;
+	char *override_label;
+};
+
+struct oc_form_opt_select {
+	struct oc_form_opt form;
+	int nr_choices;
+	struct oc_choice choices[0];
+};
+
+/* All char * fields are static, owned by the XML parser */
+struct oc_auth_form {
+	char *banner;
+	char *message;
+	char *error;
+	char *auth_id;
+	char *method;
+	char *action;
+	struct oc_form_opt *opts;
+};
+
+/****************************************************************************/
+
 struct pkt {
 	int type;
 	int len;
@@ -41,7 +91,7 @@ struct pkt {
 	unsigned char hdr[8];
 	unsigned char data[];
 };
-	
+
 struct vpn_option {
 	char *option;
 	char *value;
@@ -72,6 +122,11 @@ struct split_include {
 #define RECONNECT_INTERVAL_MIN	10
 #define RECONNECT_INTERVAL_MAX	100
 
+#define CERT_TYPE_UNKNOWN	0
+#define CERT_TYPE_PEM		1
+#define CERT_TYPE_PKCS12	2
+#define CERT_TYPE_TPM		3
+
 struct openconnect_info {
 	char *redirect_url;
 
@@ -85,14 +140,17 @@ struct openconnect_info {
 	char *urlpath;
 	const char *cert;
 	const char *sslkey;
-	int tpm;
-	char *tpmpass;
+	int cert_type;
+	char *cert_password;
 	const char *cafile;
+	const char *servercert;
 	const char *xmlconfig;
 	char xmlsha1[(SHA_DIGEST_LENGTH * 2) + 1];
 	char *username;
 	char *password;
+	char *authgroup;
 	int nopasswd;
+	char *dtls_ciphers;
 
 	char *cookie;
 	struct vpn_option *cookies;
@@ -123,6 +181,7 @@ struct openconnect_info {
 	unsigned char dtls_session_id[32];
 	unsigned char dtls_secret[48];
 
+	char *dtls_cipher;
 	char *vpnc_script;
 	int script_tun;
 	char *ifname;
@@ -133,7 +192,9 @@ struct openconnect_info {
 	const char *vpn_dns[3];
 	const char *vpn_nbns[3];
 	const char *vpn_domain;
+	const char *vpn_proxy_pac;
 	struct split_include *split_includes;
+	struct split_include *split_excludes;
 
 	int select_nfds;
 	fd_set select_rfds;
@@ -160,6 +221,7 @@ struct openconnect_info {
 
 	int (*validate_peer_cert) (struct openconnect_info *vpninfo, X509 *cert);
 	int (*write_new_config) (struct openconnect_info *vpninfo, char *buf, int buflen);
+	int (*process_auth_form) (struct openconnect_info *vpninfo, struct oc_auth_form *form);
 
 	void __attribute__ ((format(printf, 3, 4)))
 	(*progress) (struct openconnect_info *vpninfo, int level, const char *fmt, ...);
@@ -181,6 +243,8 @@ struct openconnect_info {
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
 
+/****************************************************************************/
+
 /* tun.c */
 int setup_tun(struct openconnect_info *vpninfo);
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
@@ -203,9 +267,9 @@ int  __attribute__ ((format (printf, 2, 3)))
 int openconnect_SSL_gets(SSL *ssl, char *buf, size_t len);
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo);
-
-/* main.c */
-extern int verbose;
+int get_cert_fingerprint(X509 *cert, char *buf);
+void report_ssl_errors(struct openconnect_info *vpninfo);
+int passphrase_from_fsid(struct openconnect_info *vpninfo);
 
 /* mainloop.c */
 int vpn_add_pollfd(struct openconnect_info *vpninfo, int fd, short events);
@@ -220,6 +284,10 @@ extern int killed;
 /* xml.c */
 int config_lookup_host(struct openconnect_info *vpninfo, const char *host);
 
+/* auth.c */
+int parse_xml_response(struct openconnect_info *vpninfo, char *response,
+		       char *request_body, int req_len);
+
 /* http.c */
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo);
 char *openconnect_create_useragent(char *base);
