@@ -22,23 +22,29 @@
  *   Boston, MA 02110-1301 USA
  */
 
-#include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
+#include <string.h>
 #ifdef __linux__
 #include <linux/if_tun.h>
 #endif
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <netinet/in_systm.h>
 #include <netinet/in.h>
+#include <netinet/ip.h>
 #include <net/if.h>
 #include <arpa/inet.h>
 #include <errno.h>
 
 #include "openconnect.h"
 
+#ifdef BSD
+#define TUN_HAS_AF_PREFIX 1
+#endif
+
 static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 {
 	struct ifreq ifr;
@@ -58,7 +64,7 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 		if (ioctl(net_fd, SIOCGIFFLAGS, &ifr) < 0)
 			perror("SIOCGIFFLAGS");
 
-		ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT; 
+		ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT;
 		if (ioctl(net_fd, SIOCSIFFLAGS, &ifr) < 0)
 			perror("SIOCSIFFLAGS");
 
@@ -84,8 +90,8 @@ static int setenv_int(const char *opt, int value)
 	return setenv(opt, buf, 1);
 }
 
-static int process_split_include(struct openconnect_info *vpninfo,
-				 char *route, int *nr_incs)
+static int process_split_xxclude(struct openconnect_info *vpninfo,
+				 char *in_ex, char *route, int *nr_incs)
 {
 	struct in_addr addr;
 	int masklen;
@@ -96,8 +102,8 @@ static int process_split_include(struct openconnect_info *vpninfo,
 	if (!slash) {
 	badinc:
 		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Discard bad split include: \"%s\"\n",
-				  route);
+				  "Discard bad split %sclude: \"%s\"\n",
+				  in_ex, route);
 		return -EINVAL;
 	}
 
@@ -108,7 +114,7 @@ static int process_split_include(struct openconnect_info *vpninfo,
 	}
 
 	envname[79] = 0;
-	snprintf(envname, 79, "CISCO_SPLIT_INC_%d_ADDR", *nr_incs);
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_ADDR", in_ex, *nr_incs);
 	setenv(envname, route, 1);
 
 	/* Put it back how we found it */
@@ -117,7 +123,7 @@ static int process_split_include(struct openconnect_info *vpninfo,
 	if (!inet_aton(slash+1, &addr))
 		goto badinc;
 
-	snprintf(envname, 79, "CISCO_SPLIT_INC_%d_MASK", *nr_incs);
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASK", in_ex, *nr_incs);
 	setenv(envname, slash+1, 1);
 
 	for (masklen = 0; masklen < 32; masklen++) {
@@ -125,8 +131,8 @@ static int process_split_include(struct openconnect_info *vpninfo,
 			break;
 	}
 	masklen = 32 - masklen;
-		    
-	snprintf(envname, 79, "CISCO_SPLIT_INC_%d_MASKLEN", *nr_incs);
+
+	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASKLEN", in_ex, *nr_incs);
 	setenv_int(envname, masklen);
 
 	(*nr_incs)++;
@@ -147,6 +153,30 @@ static int appendenv(const char *opt, const char *new)
 	return setenv(opt, buf, 1);
 }
 
+static void setenv_cstp_opts(struct openconnect_info *vpninfo)
+{
+	char *env_buf;
+	int buflen = 0;
+	int bufofs = 0;
+	struct vpn_option *opt;
+
+	for (opt = vpninfo->cstp_options; opt; opt = opt->next)
+		buflen += 2 + strlen(opt->option) + strlen(opt->value);
+
+	env_buf = malloc(buflen + 1);
+	if (!env_buf)
+		return;
+
+	env_buf[buflen] = 0;
+
+	for (opt = vpninfo->cstp_options; opt; opt = opt->next)
+		bufofs += snprintf(env_buf + bufofs, buflen - bufofs,
+				   "%s=%s\n", opt->option, opt->value);
+
+	setenv("CISCO_CSTP_OPTIONS", env_buf, 1);
+	free(env_buf);
+}
+
 static void set_script_env(struct openconnect_info *vpninfo)
 {
 	struct sockaddr_in *sin = (void *)vpninfo->peer_addr;
@@ -156,12 +186,13 @@ static void set_script_env(struct openconnect_info *vpninfo)
 	setenv("reason", "connect", 1);
 	unsetenv("CISCO_BANNER");
 	unsetenv("CISCO_SPLIT_INC");
+	unsetenv("CISCO_SPLIT_EXC");
 
 	setenv_int("INTERNAL_IP4_MTU", vpninfo->mtu);
 
 	setenv("INTERNAL_IP4_ADDRESS", vpninfo->vpn_addr, 1);
 	setenv("INTERNAL_IP4_NETMASK", vpninfo->vpn_netmask, 1);
-	
+
 	if (vpninfo->vpn_dns[0])
 		setenv("INTERNAL_IP4_DNS", vpninfo->vpn_dns[0], 1);
 	else
@@ -184,19 +215,32 @@ static void set_script_env(struct openconnect_info *vpninfo)
 		setenv("CISCO_DEF_DOMAIN", vpninfo->vpn_domain, 1);
 	else unsetenv ("CISCO_DEF_DOMAIN");
 
+	if (vpninfo->vpn_proxy_pac)
+		setenv("CISCO_PROXY_PAC", vpninfo->vpn_proxy_pac, 1);
+
 	if (vpninfo->split_includes) {
 		struct split_include *this = vpninfo->split_includes;
 		int nr_split_includes = 0;
 
 		while (this) {
-			process_split_include(vpninfo, this->route,
+			process_split_xxclude(vpninfo, "IN", this->route,
 					      &nr_split_includes);
 			this = this->next;
 		}
 		setenv_int("CISCO_SPLIT_INC", nr_split_includes);
-	}			
-			
-			
+	}
+	if (vpninfo->split_excludes) {
+		struct split_include *this = vpninfo->split_excludes;
+		int nr_split_excludes = 0;
+
+		while (this) {
+			process_split_xxclude(vpninfo, "EX", this->route,
+					      &nr_split_excludes);
+			this = this->next;
+		}
+		setenv_int("CISCO_SPLIT_EXC", nr_split_excludes);
+	}
+	setenv_cstp_opts(vpninfo);
 }
 
 static int script_config_tun(struct openconnect_info *vpninfo)
@@ -216,7 +260,6 @@ static int script_config_tun(struct openconnect_info *vpninfo)
 /* Set up a tuntap device. */
 int setup_tun(struct openconnect_info *vpninfo)
 {
-	struct ifreq ifr;
 	int tun_fd;
 
 	if (vpninfo->script_tun) {
@@ -244,6 +287,8 @@ int setup_tun(struct openconnect_info *vpninfo)
 		vpninfo->ifname = "(script)";
 	} else {
 #ifdef IFF_TUN /* Linux */
+		struct ifreq ifr;
+
 		tun_fd = open("/dev/net/tun", O_RDWR);
 		if (tun_fd < 0) {
 			vpninfo->progress(vpninfo, PRG_ERR,
@@ -268,7 +313,7 @@ int setup_tun(struct openconnect_info *vpninfo)
 #else /* BSD et al have /dev/tun$x devices */
 		static char tun_name[80];
 		int i;
-		for (i=0; i < 255; i++) {
+		for (i = 0; i < 255; i++) {
 			sprintf(tun_name, "/dev/tun%d", i);
 			tun_fd = open(tun_name, O_RDWR);
 			if (tun_fd >= 0)
@@ -284,14 +329,14 @@ int setup_tun(struct openconnect_info *vpninfo)
 			script_config_tun(vpninfo);
 			/* We have to set the MTU for ourselves, because the script doesn't */
 			local_config_tun(vpninfo, 1);
-		} else 
+		} else
 			local_config_tun(vpninfo, 0);
 	}
 
 	fcntl(tun_fd, F_SETFD, FD_CLOEXEC);
 
 	vpninfo->tun_fd = tun_fd;
-	
+
 	if (vpninfo->select_nfds <= tun_fd)
 		vpninfo->select_nfds = tun_fd + 1;
 
@@ -304,13 +349,31 @@ int setup_tun(struct openconnect_info *vpninfo)
 
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 {
-	char buf[2000];
+	unsigned char buf[2000];
 	int len;
 	int work_done = 0;
 
 	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
 		while ((len = read(vpninfo->tun_fd, buf, sizeof(buf))) > 0) {
-			if (queue_new_packet(&vpninfo->outgoing_queue, AF_INET, buf, len))
+			unsigned char *pkt = buf;
+			int type;
+#ifdef __OpenBSD__
+			type = ntohl(*(int *)buf);
+			if (type != AF_INET && type != AF_INET6)
+				continue;
+			pkt += 4;
+			len -= 4;
+#else
+			struct ip *iph = (void *)buf;
+			if (iph->ip_v == 6)
+				type = AF_INET6;
+			else if (iph->ip_v == 4)
+				type = AF_INET;
+			else
+				continue;
+#endif
+			if (queue_new_packet(&vpninfo->outgoing_queue, type,
+					     pkt, len))
 				break;
 
 			work_done = 1;
@@ -329,8 +392,18 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	   the no-longer-full situation, so let's not bother. */
 	while (vpninfo->incoming_queue) {
 		struct pkt *this = vpninfo->incoming_queue;
+		unsigned char *data = this->data;
+		int len = this->len;
+
 		vpninfo->incoming_queue = this->next;
-		if (write(vpninfo->tun_fd, this->data, this->len) < 0 &&
+
+#ifdef __OpenBSD__
+		data -= 4;
+		len += 4;
+		*(int *)data = htonl(this->type);
+#endif
+
+		if (write(vpninfo->tun_fd, data, len) < 0 &&
 		    errno == ENOTCONN) {
 			vpninfo->quit_reason = "Client connection terminated";
 			return 1;
